##############################################################################
#
# Copyright (c) 2019 Nexedi SA and Contributors. All Rights Reserved.
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsibility of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# guarantees and support are strongly adviced to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
##############################################################################
from __future__ import unicode_literals

import os
import textwrap
import logging
import subprocess
import tempfile
import time
import re
import shutil
from six.moves.urllib.parse import urlparse, urljoin

import pexpect
import psutil
import requests
import sqlite3

from slapos.testing.testcase import makeModuleSetUpAndTestCaseClass
from slapos.grid.svcbackend import getSupervisorRPC
from slapos.grid.svcbackend import _getSupervisordSocketPath


# Base classes
# ------------

setUpModule, SlapOSInstanceTestCase = makeModuleSetUpAndTestCaseClass(
    os.path.abspath(
        os.path.join(os.path.dirname(__file__), '..', 'software.cfg')))


class TheiaTestCase(SlapOSInstanceTestCase):
    # Theia uses unix sockets, so it needs short paths.
  __partition_reference__ = 'T'

  @classmethod
  def getPartitionIdWhere(cls, software_url, software_type):
    for computer_partition in cls.slap.computer.getComputerPartitionList():
      _software = computer_partition.getSoftwareRelease()._software_release
      _type = computer_partition.getType()
      if _software == software_url and _type == software_type:
        return computer_partition.getId()
    return None

  @classmethod
  def getPartitionPathWhere(cls, software_url, software_type):
    Id = cls.getPartitionIdWhere(software_url, software_type)
    if Id:
      return os.path.join(cls.slap._instance_root, Id)
    return None


class TheiaResiliencyTestCase(TheiaTestCase):

  @classmethod
  def _getLocalPath(cls, relative_path):
    partition_root = cls.getPartitionPathWhere(cls.getSoftwareURL(), 'export')
    absolute_path = os.path.join(partition_root, relative_path)
    return absolute_path

  @classmethod
  def _doSlaposCommand(cls, command):
    partition_root = cls.getPartitionPathWhere(cls.getSoftwareURL(), 'export')
    slapos_bin = os.path.join(partition_root, 'srv', 'runner', 'bin', 'slapos')
    output = subprocess.check_output((slapos_bin, ) + command)
    return output

  def _supplySoftwareRelease(self, software_url):
    self._doSlaposCommand(('supply', software_url, 'slaprunner'))

  def _waitForSoftwareBuild(self):
    self._doSlaposCommand(('node', 'software'))

  def _requestInstance(self, instance_name, software_url):
    self._doSlaposCommand(('request', instance_name, software_url))

  def _waitForInstanceDeploy(self):
    self._doSlaposCommand(('node', 'instance'))

  def _getTakeoverUrlAndPassword(self, scope="theia-1"):
    parameter_dict = self.computer_partition.getConnectionParameterDict()
    takeover_url = parameter_dict["takeover-%s-url" % scope]
    takeover_password = parameter_dict["takeover-%s-password" % scope]
    return takeover_url, takeover_password

  def _getTakeoverState(self, takeover_url):
    def getTakeoverPageContent():
      resp = requests.get(takeover_url, verify=True)
      self.assertEqual(requests.codes.ok, resp.status_code)
      return resp.text

    takeover_page_content = getTakeoverPageContent()
    if "<b>Last valid backup:</b> No backup downloaded yet, takeover should not happen now." in takeover_page_content:
      return "nothing"
    elif "<b>Importer script(s) of backup in progress:</b> True" in takeover_page_content:
      return "ongoing"
    return "ready"

  def _doTakeover(self, takeover_url, takeover_password):
    resp = requests.get(
      "%s?password=%s" % (takeover_url, takeover_password),
      verify=True
    )
    self.assertEqual(requests.codes.ok, resp.status_code)
    self.assertNotIn("Error", resp.text, "An Error occured: %s" % resp.text)
    self.assertIn("Success", resp.text, "'Success' not in '%s'" % resp.text)
    return resp.text


# Utils
# -----

class SQLiteDB(object):
  def __init__(self, sqlitedb_file):
    self.sqlitedb_file = sqlitedb_file

  def select(self, fields, table, where={}):
    connection = sqlite3.connect(self.sqlitedb_file)

    def dict_factory(cursor, row):
      d = {}
      for idx, col in enumerate(cursor.description):
        d[col[0]] = row[idx]
      return d
    connection.row_factory = dict_factory
    cursor = connection.cursor()

    condition = " AND ".join("%s='%s'" % (k, v) for k, v in where.items())
    cursor.execute(
      "SELECT %s FROM %s%s"
      % (
        ", ".join(fields),
        table,
        " WHERE %s" % condition if where else "",
      )
    )
    return cursor.fetchall()


# Tests
# -----

class TestTheiaInterface(TheiaTestCase):
  def setUp(self):
    self.connection_parameters = self.computer_partition.getConnectionParameterDict()

  def test_http_get(self):
    resp = requests.get(self.connection_parameters['url'], verify=False)
    self.assertEqual(requests.codes.unauthorized, resp.status_code)

    # with login/password, this is allowed
    parsed_url = urlparse(self.connection_parameters['url'])
    authenticated_url = parsed_url._replace(
        netloc='{}:{}@[{}]:{}'.format(
            self.connection_parameters['username'],
            self.connection_parameters['password'],
            parsed_url.hostname,
            parsed_url.port,
        )).geturl()
    resp = requests.get(authenticated_url, verify=False)
    self.assertEqual(requests.codes.ok, resp.status_code)

    # there's a public folder to serve file
    with open('{}/srv/frontend-static/public/test_file'.format(
        self.computer_partition_root_path), 'w') as f:
      f.write("hello")
    resp = requests.get(urljoin(authenticated_url, '/public/'), verify=False)
    self.assertIn('test_file', resp.text)
    resp = requests.get(
        urljoin(authenticated_url, '/public/test_file'), verify=False)
    self.assertEqual('hello', resp.text)

    # there's a (not empty) favicon
    resp = requests.get(
        urljoin(authenticated_url, '/favicon.ico'), verify=False)
    self.assertEqual(requests.codes.ok, resp.status_code)
    self.assertTrue(resp.raw)

    # there is a CSS referencing fonts
    css_text = requests.get(urljoin(authenticated_url, '/css/slapos.css'), verify=False).text
    css_urls = re.findall(r'url\([\'"]+([^\)]+)[\'"]+\)', css_text)
    self.assertTrue(css_urls)
    # and fonts are served
    for url in css_urls:
      resp = requests.get(urljoin(authenticated_url, url), verify=False)
      self.assertEqual(requests.codes.ok, resp.status_code)
      self.assertTrue(resp.raw)

  def test_theia_slapos(self):
    # Make sure we can use the shell and the integrated slapos command
    process = pexpect.spawnu(
        '{}/bin/theia-shell'.format(self.computer_partition_root_path),
        env={'HOME': self.computer_partition_root_path})

    # use a large enough terminal so that slapos proxy show table fit in the screen
    process.setwinsize(5000, 5000)

    # log process output for debugging
    logger = logging.getLogger('theia-shell')
    class DebugLogFile:
      def write(self, msg):
        logger.info("output from theia-shell: %s", msg)
      def flush(self):
        pass
    process.logfile = DebugLogFile()

    process.expect_exact('Standalone SlapOS for computer `slaprunner` activated')

    # try to supply and install a software to check that this slapos is usable
    process.sendline(
        'slapos supply https://lab.nexedi.com/nexedi/slapos/raw/1.0.144/software/helloworld/software.cfg slaprunner'
    )
    process.expect(
        'Requesting software installation of https://lab.nexedi.com/nexedi/slapos/raw/1.0.144/software/helloworld/software.cfg...'
    )

    # we pipe through cat to disable pager and prevent warnings like
    # WARNING: terminal is not fully functional
    process.sendline('slapos proxy show | cat')
    process.expect(
        'https://lab.nexedi.com/nexedi/slapos/raw/1.0.144/software/helloworld/software.cfg'
    )

    process.sendline('slapos node software')
    process.expect(
        'Installing software release https://lab.nexedi.com/nexedi/slapos/raw/1.0.144/software/helloworld/software.cfg'
    )
    # interrupt this, we don't want to actually wait for software installation
    process.sendcontrol('c')

    process.terminate()
    process.wait()

  def test_theia_shell_execute_tasks(self):
    # shell needs to understand -c "command" arguments for theia tasks feature
    test_file = '{}/test file'.format(self.computer_partition_root_path)
    subprocess.check_call([
        '{}/bin/theia-shell'.format(self.computer_partition_root_path),
        '-c',
        'touch "{}"'.format(test_file)
    ])
    self.assertTrue(os.path.exists(test_file))


class TestTheiaEmbeddedSlapOSShutdown(TheiaTestCase):
  def test_stopping_instance_stops_embedded_slapos(self):
    embedded_slapos_supervisord_socket = _getSupervisordSocketPath(
        os.path.join(
            self.computer_partition_root_path,
            'srv',
            'runner',
            'instance',
        ), self.logger)

    # Wait a bit for this supervisor to be started.
    for _ in range(20):
      if os.path.exists(embedded_slapos_supervisord_socket):
        break
      time.sleep(1)

    # get the pid of the supervisor used to manage instances
    with getSupervisorRPC(embedded_slapos_supervisord_socket) as embedded_slapos_supervisor:
      embedded_slapos_process = psutil.Process(embedded_slapos_supervisor.getPID())

    # Stop theia's services
    with self.slap.instance_supervisor_rpc as instance_supervisor:
      process_info, = [
          p for p in instance_supervisor.getAllProcessInfo()
          if p['name'].startswith('slapos-standalone-instance-')
      ]
      instance_supervisor.stopProcessGroup(process_info['group'])

    # the supervisor controlling instances is also stopped
    self.assertFalse(embedded_slapos_process.is_running())


class TestTheiaWithSR(TheiaTestCase):
  srurl = 'bogus/software.cfg'
  srtype = 'bogus'

  @classmethod
  def getInstanceParameterDict(cls):
    return {
      'embedded-sr': cls.srurl,
      'embedded-sr-type': cls.srtype,
    }

  def test(self):
    db = SQLiteDB(os.path.join(self.computer_partition_root_path, 'srv', 'runner', 'var', 'proxy.db'))
    supplied = db.select(
      fields=["*"],
      table = "software14",
      where={'url': self.srurl}
    )
    self.assertEqual(len(supplied), 1)
    requested = db.select(
      fields=["*"],
      table = "partition14",
      where={'software_release': self.srurl, 'software_type': self.srtype}
    )
    self.assertEqual(len(requested), 1)


class TestTheiaResilientInterface(TestTheiaInterface):
  instance_max_retry = 30

  @classmethod
  def getInstanceSoftwareType(cls):
    return 'resilient'

  @classmethod
  def setUpClass(cls):
    super(TestTheiaResilientInterface, cls).setUpClass()
    # Patch the computer root path to that of the export theia instance
    path = cls.getPartitionPathWhere(cls.getSoftwareURL(), 'export')
    cls.computer_partition_root_path = path


class TestTheiaResilientWithSR(TestTheiaWithSR):
  instance_max_retry = 30

  @classmethod
  def getInstanceSoftwareType(cls):
    return 'resilient'

  @classmethod
  def setUpClass(cls):
    super(TestTheiaResilientWithSR, cls).setUpClass()
    # Patch the computer root path to that of the export theia instance
    path = cls.getPartitionPathWhere(cls.getSoftwareURL(), 'export')
    cls.computer_partition_root_path = path


class TestTheiaBasicResilience(TheiaResiliencyTestCase):
  instance_max_retry = 30

  @classmethod
  def getInstanceSoftwareType(cls):
    return 'resilient'

  def test_basic_dummy_resilience(self):
    export_partition_id = self.getPartitionIdWhere(self.getSoftwareURL(), 'export')
    import_partition_id = self.getPartitionIdWhere(self.getSoftwareURL(), 'import')

    # Copy ./dummy SR in export theia ~/srv/project/dummy
    dummy_target_path = self._getLocalPath('srv/project/dummy')
    shutil.copytree('dummy', dummy_target_path)

    dummy_software_url = os.path.join(dummy_target_path, 'software.cfg')

    # Deploy dummy instance
    self._supplySoftwareRelease(dummy_software_url)
    self._waitForSoftwareBuild()
    self._requestInstance("dummy_instance", dummy_software_url)
    self._waitForInstanceDeploy()

    # Check that dummy instance was properly deployed
    relative_dummy_log_path = 'srv/runner/instance/slappart0/log.log'
    with open(self._getLocalPath(relative_dummy_log_path), 'r') as f:
      content = f.read()
    self.assertTrue(content.startswith("Hello"), content)

    # Call exporter script instead of waiting for cron job
    # XXX Accelerate cron frequency instead ?
    exporter_script = self._getLocalPath('bin/exporter')
    transaction_id = str(int(time.time()))
    subprocess.check_output((exporter_script, '--transaction-id', transaction_id))

    # Get equeue.log file in import instance
    import_partition_root = self.getPartitionPathWhere(self.getSoftwareURL(), 'import')
    importer_log = os.path.join(import_partition_root, 'var', 'log', 'equeue.log')

    def getFileContent(file_path):
      with open(file_path) as f:
        return f.read()

    # Wait for importer to start
    time.sleep(60)
    takeover_url, takeover_password = self._getTakeoverUrlAndPassword()

    for _ in range(100):
      takeover_state = self._getTakeoverState(takeover_url)
      if takeover_state == "ready":
        break
      elif takeover_state == "nothing":
          importer_log_content = getFileContent(importer_log)
          self.fail("Backup is not even started: %s" % importer_log_content)
      else:
        self.logger.info("Backup is ongoing, waiting some more")
      time.sleep(20)

    # Takeover
    self._doTakeover(takeover_url, takeover_password)

    # Wait for import instance to become export instance and new import to be allocated
    self.slap.waitForInstance(1)

    previous_computer_partition = self.computer_partition
    self.computer_partition = self.requestDefaultInstance()

    # Check that the import instance became the export instance
    new_export_partition_id = self.getPartitionIdWhere(self.getSoftwareURL(), 'export')
    self.assertNotEqual(export_partition_id, import_partition_id)
    self.assertEqual(import_partition_id, new_export_partition_id)

    # Check that there is a new import instance
    new_import_partition_id = self.getPartitionIdWhere(self.getSoftwareURL(), 'import')
    self.assertNotEqual(export_partition_id, new_import_partition_id)
    self.assertNotEqual(new_export_partition_id, new_import_partition_id)

    # Check that the data was transfered over
    with open(self._getLocalPath(relative_dummy_log_path), 'r') as f:
      content_after = f.read()

    self.assertTrue(content.startswith("Hello"), content)
    self.assertTrue(content_after.startswith("Hello"), content_after)
    self.assertIn(content, content_after, "%s not in %s" % (content, content_after))
